from typing import (
    TYPE_CHECKING,
    Optional,
)

from eth_utils.toolz import (
    partial,
)
from eth_typing import (
    ABI,
)

from web3.contract.base_contract import (
    BaseContractCaller,
)


from conflux_web3.types import (
    TxParam,
    AddressParam,
    EpochNumberParam,
)

from .function import (
    ConfluxContractFunctions,
)

if TYPE_CHECKING:
    from conflux_web3 import Web3
    

class ConfluxContractCaller(BaseContractCaller):
    def __init__(
        self,
        abi: ABI,
        w3: "Web3",
        address: AddressParam,
        transaction: Optional[TxParam] = None,
        block_identifier: EpochNumberParam = "latest_state",
        ccip_read_enabled: Optional[bool] = None,
        decode_tuples: Optional[bool] = False,
        contract_functions: Optional[ConfluxContractFunctions] = None,
    ) -> None:
        super().__init__(
            abi,
            w3,
            address, # type: ignore
            decode_tuples=decode_tuples
        )


        if self.abi:
            if transaction is None:
                transaction = {}

            if contract_functions is None:
                contract_functions = ConfluxContractFunctions(
                    abi, w3, address=address, decode_tuples=decode_tuples
                )

            self._functions = contract_functions._functions
            for fn in contract_functions.__iter__():
                caller_method = partial(
                    self.call_function,
                    fn,
                    transaction=transaction,
                    block_identifier=block_identifier,
                    ccip_read_enabled=ccip_read_enabled,
                )
                setattr(self, str(fn.abi_element_identifier), caller_method)
        

    def __call__(
        self,
        transaction: Optional[TxParam] = None,
        block_identifier: EpochNumberParam = "latest_state",
        ccip_read_enabled: Optional[bool] = None,
    ) -> "ConfluxContractCaller":
        if transaction is None:
            transaction = {}
        return type(self)(
            self.abi,
            self.w3, # type: ignore
            self.address,
            transaction=transaction,
            block_identifier=block_identifier,
            ccip_read_enabled=ccip_read_enabled,
            decode_tuples=self.decode_tuples,
        )



from typing import (
    Optional,
    cast,
    Any,
)
import itertools

from eth_utils.abi import (
    get_abi_input_names,
)
from eth_abi.codec import (
    ABICodec,
)

from eth_utils.conversions import (
    to_bytes,
)
from eth_utils.toolz import (
    curry, # type: ignore
)

from web3._utils.events import (
    get_event_abi_types_for_decoding,
)
from web3._utils.abi import (
    exclude_indexed_event_inputs,
    get_indexed_event_inputs,
    map_abi_data,
    normalize_event_input_types,
    named_tree
)
from web3.utils.abi import (
    get_event_log_topics,
)
from eth_typing import (
    ABIEvent,
)
from web3.datastructures import (
    AttributeDict
)
from web3.exceptions import (
    InvalidEventABI,
    LogTopicError,
)
from web3._utils.encoding import (
    hexstr_if_str,
)

from cfx_address.utils import (
    normalize_to
)
from conflux_web3.types import (
    EventData,
    LogReceipt,
)

def _log_entry_data_to_bytes(
    log_entry_data: Any,
):
    return hexstr_if_str(to_bytes, log_entry_data)  # type: ignore

def get_cfx_base32_normalizer(chain_id: int): # type: ignore
    return lambda type_str, hex_address: (type_str, normalize_to(hex_address, chain_id, True)) if type_str == "address" else (type_str, hex_address) # type: ignore


@curry # type: ignore
def cfx_get_event_data(
    abi_codec: ABICodec,
    event_abi: ABIEvent,
    log_entry: LogReceipt,
    chain_id: Optional[int]= None
) -> EventData:
    """
    Given an event ABI and a log entry for that event, return the decoded
    event data
    """
    log_topics = get_event_log_topics(event_abi, log_entry["topics"])
    log_topics_bytes = [_log_entry_data_to_bytes(topic) for topic in log_topics]
    log_topics_abi = get_indexed_event_inputs(event_abi)
    log_topic_normalized_inputs = normalize_event_input_types(log_topics_abi)
    log_topic_types = get_event_abi_types_for_decoding(log_topic_normalized_inputs)
    log_topic_names = get_abi_input_names(
        ABIEvent({"name": event_abi["name"], "type": "event", "inputs": log_topics_abi})
    )

    if len(log_topics_bytes) != len(log_topic_types):
        raise LogTopicError(
            f"Expected {len(log_topic_types)} log topics.  Got {len(log_topics_bytes)}"
        )

    log_data = _log_entry_data_to_bytes(log_entry["data"])
    log_data_abi = exclude_indexed_event_inputs(event_abi)
    log_data_normalized_inputs = normalize_event_input_types(log_data_abi)
    log_data_types = get_event_abi_types_for_decoding(log_data_normalized_inputs)
    log_data_names = get_abi_input_names(
        ABIEvent({"name": event_abi["name"], "type": "event", "inputs": log_data_abi})
    )

    # sanity check that there are not name intersections between the topic
    # names and the data argument names.
    duplicate_names = set(log_topic_names).intersection(log_data_names)
    if duplicate_names:
        raise InvalidEventABI(
            "The following argument names are duplicated "
            f"between event inputs: '{', '.join(duplicate_names)}'"
        )

    decoded_log_data = abi_codec.decode(log_data_types, log_data)
    normalized_log_data = map_abi_data(
        [get_cfx_base32_normalizer(chain_id)], log_data_types, decoded_log_data
    )
    named_log_data = named_tree(
        log_data_normalized_inputs,
        normalized_log_data,
    )

    decoded_topic_data = [
        abi_codec.decode([topic_type], topic_data)[0]
        for topic_type, topic_data in zip(log_topic_types, log_topics_bytes)
    ]
    normalized_topic_data = map_abi_data(
        [get_cfx_base32_normalizer(chain_id)], log_topic_types, decoded_topic_data
    )

    event_args = dict(
        itertools.chain(
            zip(log_topic_names, normalized_topic_data),
            named_log_data.items(),
        )
    )

    event_data = {
        "args": event_args,
        "event": event_abi.get("name", None),
        "logIndex": log_entry.get("logIndex", None),
        "transactionIndex": log_entry.get("transactionIndex", None),
        "transactionLogIndex": log_entry.get("transactionLogIndex", None),
        "transactionHash": log_entry.get("transactionHash", None),
        "address": log_entry["address"],
        "blockHash": log_entry.get("blockHash", None),
        "epochNumber": log_entry.get("epochNumber", None)
    }
    if isinstance(log_entry, AttributeDict):
        return cast(EventData, AttributeDict.recursive(event_data))

    return event_data
    

# events.get_event_data = conditional_func(
#     cfx_get_event_data,
#     from_cfx_condition
# )(events.get_event_data)


# modify_to_conditional_func(events.get_event_data, cfx_get_event_data, from_cfx_condition)





from typing import Any
from web3.types import RPCEndpoint
from conflux_web3._utils.rpc_abi import (
    RPC
)
from conflux_web3.types.transaction_hash import (
    TransactionHash
)

from conflux_web3.middleware.base import ConfluxWeb3Middleware
    
class PendingTransactionMiddleware(ConfluxWeb3Middleware):
    def response_processor(self, method: RPCEndpoint, response: Any):
        if method == RPC.cfx_sendTransaction or method == RPC.cfx_sendRawTransaction:
            if "result" in response:
                transaction_hash = TransactionHash(response["result"])
                transaction_hash.set_w3(self._w3)
                
                response["result"] = transaction_hash
        return response





from typing import (
    TYPE_CHECKING,
    List,
    Optional,
    cast,
)
from hexbytes import HexBytes
from eth_utils.toolz import (
    # assoc,
    curry,  # type: ignore
    merge,  # type: ignore
)

from cfx_utils.token_unit import (
    Drip
)
from cfx_utils.types import (
    CIP1559TxDict
)
from conflux_web3.exceptions import (
    NoWeb3Exception
)
from conflux_web3.types import (
    TxParam,
)
from conflux_web3._utils.cns import (
    resolve_if_cns_name,
)

if TYPE_CHECKING:
    from conflux_web3 import Web3


LEGACY_TRANSACTION_DEFAULTS = {
    "value": 0,
    "data": b"",
    "nonce": lambda w3, tx, estimate=None: w3.cfx.get_next_nonce(tx['from']), # type: ignore
    "gas": lambda w3, tx, estimate=None: estimate["gasLimit"], # type: ignore
    "storageLimit": lambda w3, tx, estimate=None: estimate["storageCollateralized"], # type: ignore
    # convert to int value
    "gasPrice": lambda w3, tx, estimate=None: w3.cfx.gas_price.to(Drip).value, # type: ignore
    "chainId": lambda w3, tx, estimate=None: w3.cfx.chain_id, # type: ignore
    "epochHeight": lambda w3, tx, estimate=None: w3.cfx.epoch_number, # type: ignore
}


@curry
def fill_transaction_defaults(w3: "Web3", transaction: TxParam) -> TxParam:
    """
    Fill the necessary fields to "send" a transaction
    Before this function is invoked, ensure 'from' field is filled
    """
    if not w3:
        raise NoWeb3Exception("A web3 object is required to fill transaction defaults, but no web3 object is passed")
    if (not transaction.get("from")) and (transaction.get("nonce", None) is None):
        raise ValueError("Transaction's 'from' field is required to fill nonce field")
    if "from" in transaction:
        transaction['from'] = resolve_if_cns_name(w3, transaction['from'])
    
    transaction_type = tell_transaction_type(transaction)
    if transaction_type == 0:
        return set_legacy_transaction_defaults(w3, transaction)
    elif transaction_type == 2:
        return set_cip1559_transaction_defaults(w3, transaction)
    else:
        raise ValueError(f"Transaction type not supported: {transaction_type}")

def tell_transaction_type(transaction: TxParam) -> int:
    if "type" in transaction:
        return int(HexBytes(transaction["type"]))
    if "gasPrice" in transaction:
        return 0
    if "maxFeePerGas" in transaction or "maxPriorityFeePerGas" in transaction:
        return 2
    return 2

def set_legacy_transaction_defaults(w3: "Web3", transaction: TxParam) -> TxParam:
    estimate = None
    if "value" not in transaction:
        transaction["value"] = 0
    if "data" not in transaction:
        transaction["data"] = b""
    if "nonce" not in transaction:
        transaction["nonce"] = w3.cfx.get_next_nonce(transaction["from"])
    if "gas" not in transaction:
        estimate = w3.cfx.estimate_gas_and_collateral(transaction)
        transaction["gas"] = estimate["gasLimit"]
    if "storageLimit" not in transaction:
        if not estimate:
            estimate = w3.cfx.estimate_gas_and_collateral(transaction)
        transaction["storageLimit"] = estimate["storageCollateralized"]
    if "gasPrice" not in transaction:
        transaction["gasPrice"] = w3.cfx.gas_price.to(Drip).value
    if "chainId" not in transaction:
        transaction["chainId"] = w3.cfx.chain_id
    if "epochHeight" not in transaction:
        transaction["epochHeight"] = w3.cfx.epoch_number
    transaction["type"] = 0
    return transaction

def set_cip1559_transaction_defaults(w3: "Web3", transaction: CIP1559TxDict) -> TxParam:
    estimate = None
    if "value" not in transaction:
        transaction["value"] = 0
    if "data" not in transaction:
        transaction["data"] = b""
    if "nonce" not in transaction:
        transaction["nonce"] = w3.cfx.get_next_nonce(transaction["from"])
    if "gas" not in transaction:
        estimate = w3.cfx.estimate_gas_and_collateral(transaction)
        transaction["gas"] = estimate["gasLimit"]
    if "storageLimit" not in transaction:
        if not estimate:
            estimate = w3.cfx.estimate_gas_and_collateral(transaction)
        transaction["storageLimit"] = estimate["storageCollateralized"]
    if "maxPriorityFeePerGas" not in transaction:
        transaction["maxPriorityFeePerGas"] = w3.cfx.max_priority_fee.to(Drip).value
    if "maxFeePerGas" not in transaction:
        base_fee = w3.cfx.get_block("latest_state")["baseFeePerGas"] * 2
        transaction["maxFeePerGas"] = (base_fee + Drip(transaction["maxPriorityFeePerGas"])).to(Drip).value
    if "chainId" not in transaction:
        transaction["chainId"] = w3.cfx.chain_id
    if "epochHeight" not in transaction:
        transaction["epochHeight"] = w3.cfx.epoch_number
    transaction["type"] = 2
    return transaction


# a fork from web3.middleware.names
from typing import (
    TYPE_CHECKING,
)

from web3._utils.rpc_abi import (
    abi_request_formatters,
)
from web3.middleware.formatting import (
    FormattingMiddlewareBuilder,
)

from conflux_web3.types import (
    Middleware,
)
from conflux_web3._utils.normalizers import (
    abi_cns_resolver,
)
from conflux_web3._utils.rpc_abi import (
    RPC_ABIS,
)

if TYPE_CHECKING:
    from conflux_web3 import Web3  # noqa: F401


def name_to_address_middleware(w3: "Web3") -> Middleware:
    normalizers = [
        abi_cns_resolver(w3), # type: ignore
    ]
    return FormattingMiddlewareBuilder.build(
        request_formatters=abi_request_formatters(normalizers, RPC_ABIS)  # type: ignore
    )



from typing import Optional, Sequence, Type
import time
import requests
from pydantic import BaseModel
from web3.providers.rpc import HTTPProvider as OriHTTPProvider
from web3.types import RPCEndpoint
from web3._utils.empty import (
    Empty,
)


REQUEST_RETRY_DISALLOWLIST = [
    "cfx_sendTransaction",
    "cfx_sendRawTransaction",
]

def check_if_retry_on_failure(
    method: RPCEndpoint,
    allowlist: Optional[Sequence[str]] = None,
    disallowlist: Optional[Sequence[str]] = None,
) -> bool:
    if allowlist != None and disallowlist != None:
        raise ValueError("allowlist and disallowlist cannot be used together")
    elif allowlist is None and disallowlist is None:
        disallowlist = REQUEST_RETRY_DISALLOWLIST
        
    if allowlist is not None:
        if method in allowlist or method.split("_")[0] in allowlist:
            return True
        else:
            return False
    if disallowlist is not None:
        if method in disallowlist or method.split("_")[0] in disallowlist:
            return False
        else:
            return True
    
    # Should never reach here
    raise ValueError("allowlist or disallowlist must be provided")


class ConfluxExceptionRetryConfiguration(BaseModel):
    errors: Sequence[Type[BaseException]]
    retries: int
    backoff_factor: float
    method_allowlist: Optional[Sequence[str]]
    method_disallowlist: Optional[Sequence[str]]

    def __init__(
        self,
        errors: Sequence[Type[BaseException]] = None,
        retries: int = 5,
        backoff_factor: float = 0.125,
        method_allowlist: Optional[Sequence[str]] = None,
        method_disallowlist: Optional[Sequence[str]] = None,
    ):
        super().__init__(
            errors=errors,
            retries=retries,
            backoff_factor=backoff_factor,
            method_allowlist=method_allowlist,
            method_disallowlist=method_disallowlist,
        )

class HTTPProvider(OriHTTPProvider):
    @property
    def exception_retry_configuration(self) -> ConfluxExceptionRetryConfiguration:
        if isinstance(self._exception_retry_configuration, Empty):
            self._exception_retry_configuration = ConfluxExceptionRetryConfiguration(
                errors=(
                    ConnectionError,
                    requests.HTTPError,
                    requests.Timeout,
                )
            )
        return self._exception_retry_configuration
    
    def _make_request(self, method: RPCEndpoint, request_data: bytes) -> bytes:
        if (
            self.exception_retry_configuration is not None
            and check_if_retry_on_failure(
                method, self.exception_retry_configuration.method_allowlist
            )
        ):
            for i in range(self.exception_retry_configuration.retries):
                try:
                    return self._request_session_manager.make_post_request(
                        self.endpoint_uri, request_data, **self.get_request_kwargs()
                    )
                except tuple(self.exception_retry_configuration.errors) as e:
                    if i < self.exception_retry_configuration.retries - 1:
                        time.sleep(
                            self.exception_retry_configuration.backoff_factor * 2**i
                        )
                        continue
                    else:
                        raise e
            return None
        else:
            return self._request_session_manager.make_post_request(
                self.endpoint_uri, request_data, **self.get_request_kwargs()
            )
    



from typing import (
    TYPE_CHECKING, Any, Union
)
from web3._utils.ens import (
    is_ens_name
)
from web3._utils.empty import (
    empty
)
from web3.exceptions import (
    NameNotFound
)
from cfx_address.utils import (
    is_valid_base32
)
from cfx_address import (
    Base32Address
)
from conflux_web3.exceptions import (
    NameServiceNotSet
)

if TYPE_CHECKING:
    from conflux_web3 import Web3

def is_cns_name(value: Any) -> bool:
    if is_valid_base32(value):
        return False
    else:
        return is_ens_name(value) and "." in value
    
def validate_cns_existence(w3: "Web3"):
    if w3.cns is empty:
        raise NameServiceNotSet("Web3's name service is not set")

def resolve_if_cns_name(w3: "Web3", cns_name: str) -> Union[str, Base32Address]:
    if is_cns_name(cns_name):
        validate_cns_existence(w3)
        addr = w3.cns.address(cns_name)
        if addr:
            return addr
        raise NameNotFound(f"{cns_name} resolve failed")
    return cns_name



from conflux_web3.exceptions import DisabledException

disabled_method_list = [
    "get_proof",
    # "get_block",
    "get_block_transaction_count",
    "get_uncle_count",
    "get_uncle_by_block",
    "get_raw_transaction",
    "get_raw_transaction_by_block",
    "get_transaction_by_block",
    # "get_transaction_count",
    # "replace_transaction",
    # "modify_transaction"
    # "estimate_gas",
    "submit_hashrate",
    "submit_work",
]

disabled_property_list = [
    "syncing",
    "coinbase",
    "mining",
    "hashrate",
    "block_number",
    "get_work",
]

disabled_list = disabled_method_list + disabled_property_list



from typing import (
    TYPE_CHECKING,
    Optional,
    Sequence,
    Tuple,
    cast,
    Collection,
)

from ens.utils import (
    default,
)
from ens.constants import (
    ACCEPTABLE_STALE_HOURS,
)

from web3.exceptions import (
    Web3ValueError,
)
from web3.middleware.stalecheck import StalecheckMiddlewareBuilder
from cfx_address import (
    Base32Address
)

if TYPE_CHECKING:
    from conflux_web3 import Web3 as _Web3  # noqa: F401
    from web3.providers.base import (  # noqa: F401
        # AsyncBaseProvider,
        BaseProvider,
    )
    from conflux_web3.types import (
        Middleware
    )

def init_web3(
    provider: "BaseProvider" = cast("BaseProvider", default),
    middlewares: Optional[Sequence[Tuple["Middleware", str]]] = None,
    default_account: Optional[Base32Address] = None
) -> "_Web3":
    from conflux_web3 import Web3 as Web3Main
    from conflux_web3.client import ConfluxClient as CfxMain

    if provider is default:
        w3 = Web3Main(ens=None, modules={"cfx": (CfxMain)})   # type: ignore
    else:
        w3 = Web3Main(provider, middlewares, ens=None, modules={"cfx": (CfxMain)})  # type: ignore
    if default_account:
        w3.cfx._default_account = default_account
    return customize_web3(w3)

def build_stalecheck_middleware(allowable_delay: int, skip_stalecheck_for_methods: Collection[str]):
    def inner(w3):
        if allowable_delay <= 0:
            raise Web3ValueError(
                "You must set a positive allowable_delay in seconds for this middleware"
            )
        middleware = StalecheckMiddlewareBuilder(w3)
        middleware.allowable_delay = allowable_delay
        middleware.skip_stalecheck_for_methods = skip_stalecheck_for_methods
        middleware.cache = {"latest": None}
        return middleware
    return inner
    

def customize_web3(w3: "_Web3") -> "_Web3":
    if w3.middleware_onion.get("name_to_address"):
        w3.middleware_onion.remove("name_to_address")

    if not w3.middleware_onion.get("stalecheck"):
        w3.middleware_onion.add(
            build_stalecheck_middleware(ACCEPTABLE_STALE_HOURS * 3600, ["cfx_getBlockByEpochNumber", "cfx_getStatus"]), name="stalecheck"
        )
    return w3

# is_none_or_base_32_zero_address is implemented in _web3_hook
# def is_none_or_base_32_zero_address(addr) -> bool:
#     ...


import functools
import sys
from typing import (
    Any,
    Callable,
    NoReturn,
    Optional,
    TypeVar,
    Union,
    overload,
)
from typing_extensions import (
    ParamSpec,
)

from conflux_web3.exceptions import (
    DisabledException
)

if sys.version_info < (3, 8):
    from cached_property import cached_property
else:
    from functools import cached_property

T = TypeVar("T")
P = ParamSpec("P")


def temp_alter_module_variable(module: Any, varname: str, new_var: Any, condition: Callable[..., bool]):
    """temporarily alters module.varname to new_var before func is called, and recover the change afterwards

    Args:
        module (Any): the module to be temporarily changed
        varname (str): varname of the var to be changed
        new_var (Any): _description_
        condition (Callable[..., Boolean]): _description_
    """
    def inner(func: Callable[P, T]) -> Callable[P, T]:
        def wrapper(*args: P.args, **kwargs: P.kwargs):
            if condition(*args, **kwargs):
                cache = getattr(module, varname)
                module.__setattr__(varname, new_var)
                result = func(*args, **kwargs)
                module.__setattr__(varname, cache)
                return result
            return func(*args, **kwargs)
        return wrapper
    return inner

@overload
def use_instead(func: None=None, *, origin: str="This web3.eth api", substitute: Optional[str]=None) -> Callable[..., Callable[..., NoReturn]]: ...

@overload
def use_instead(func: Callable[..., Any], *, origin: str="This web3.eth api", substitute: Optional[str]=None) -> Callable[..., NoReturn]: ...

def use_instead(
    func: Optional[Callable[..., Any]]=None, *, origin: str="This web3.eth api", substitute: Optional[str]=None
) -> Union[Callable[..., Callable[..., NoReturn]], Callable[..., NoReturn]]:
    if func is None:
        return functools.partial(use_instead, origin=origin, substitute=substitute) # type: ignore
    
    @functools.wraps(func)
    def inner(*args: Any, **kwargs: Any) -> NoReturn:
        if substitute is None:
            raise DisabledException(f"{origin} is not valid in Conflux Network."
                            "Check https://developer.confluxnetwork.org/conflux-doc/docs/json_rpc/#migrating-from-ethereum-json-rpc for more information")
        else:
            raise DisabledException(f"{origin} is not valid in Conflux Network, "
                            f"use {substitute} instead")
    return inner

__all__ = [
    "cached_property",
    "use_instead",
    "temp_alter_module_variable",
]



# import functools
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Sequence,
)
import json
from hexbytes import HexBytes

from eth_typing.encoding import HexStr
from eth_utils.hexadecimal import encode_hex
from eth_utils.conversions import to_hex

from eth_typing import ABIFunction
from eth_utils.abi import (
    get_abi_input_types,
)
from web3._utils import contracts
from web3.utils.abi import (
    check_if_arguments_can_be_encoded,
)
from web3._utils.abi import (
    map_abi_data,
)
from web3._utils.normalizers import (
    abi_bytes_to_bytes,
    abi_string_to_text,
)

from conflux_web3._utils.normalizers import (
    abi_cns_resolver
)

if TYPE_CHECKING:
    from conflux_web3 import Web3

# this api is used to hook web3._utils.contracts.encode_abi
# hook is activated in _web3_hook
def cfx_encode_abi(
    web3: "Web3", abi: ABIFunction, arguments: Sequence[Any], data: Optional[HexStr] = None
) -> HexStr:
    """
    do what encode_abi does except for normalizers
    """
    argument_types = get_abi_input_types(abi)

    if not check_if_arguments_can_be_encoded(abi, *arguments, abi_codec=web3.codec):
        raise TypeError(
            "One or more arguments could not be encoded to the necessary "
            "ABI type.  Expected types are: {0}".format(
                ', '.join(argument_types),
            )
        )

    # abi_ens_resolver and abi_address_to_hex are eliminated
    normalizers = [
        abi_cns_resolver(web3), # type: ignore
        # abi_address_to_hex,
        abi_bytes_to_bytes,
        abi_string_to_text,
    ]
    normalized_arguments = map_abi_data(
        normalizers,
        argument_types,
        arguments,
    )
    encoded_arguments = web3.codec.encode(
        argument_types,
        normalized_arguments,
    )

    if data:
        return to_hex(HexBytes(data) + encoded_arguments)
    else:
        return encode_hex(encoded_arguments)

prepare_transaction = contracts.prepare_transaction



from typing import (
    Any,
    Callable,
    Dict,
    Union,
    Type,
    get_type_hints
)
from typing_extensions import (
    TypedDict,
)
from hexbytes import HexBytes

from eth_typing import (
    Hash32
)
from web3.datastructures import AttributeDict
from web3.types import (
    RPCEndpoint,
)
from web3._utils.method_formatters import (
    combine_formatters,
    # STANDARD_NORMALIZERS,
    to_hexbytes,
    to_hex_if_integer,
    to_integer_if_hex,
    is_not_null,
    apply_list_to_array_formatter,
)
from web3._utils.abi import (
    is_length
)
from web3._utils.normalizers import (
    abi_bytes_to_hex,
    abi_int_to_hex,
    abi_string_to_hex,
)
from web3.module import Module
from web3._utils.formatters import (
    is_array_of_dicts,
    is_array_of_strings,
    remove_key_if,
)
from web3._utils.rpc_abi import abi_request_formatters

from eth_utils.types import (
    is_dict
)

from eth_utils.toolz import (
    complement, # type: ignore
    compose,  # type: ignore
    # curried,
    partial, # type: ignore
)
from eth_utils.curried import (
    apply_formatter_at_index, # type: ignore
    apply_formatters_to_sequence,# type: ignore
    apply_formatters_to_dict,
    apply_formatter_if,
    apply_one_of_formatters,
)
from web3._utils.blocks import is_hex_encoded_block_hash as is_hash32_str

from cfx_address import (
    Base32Address
)
from conflux_web3._utils.rpc_abi import (
    RPC_ABIS,
    RPC
)
from conflux_web3.middleware.pending import (
    TransactionHash
)
from conflux_web3.types import (
    Drip,
    CollateralInfo,
)

STANDARD_NORMALIZERS = [
    abi_bytes_to_hex,
    abi_int_to_hex,
    abi_string_to_hex,
    # we only accept base32 format address
    # abi_address_to_hex, \
]

def to_hash32(val: Union[str, int, bytes], variable_length: bool=False):
    return to_hexbytes(32, val, variable_length)

def from_trust_to_base32(val: str):
    return Base32Address(val, _from_trust=True)

def from_hex_to_drip(val: Any):
    return Drip(val, 16)

transaction_param_formatter = compose(
    remove_key_if('to', lambda txn: txn['to'] in {'', b'', None}),  # type: ignore
    remove_key_if('gasPrice', lambda txn: txn['gasPrice'] in {'', b'', None}),  # type: ignore
    # transaction_request_formatter,
)

ABI_REQUEST_FORMATTERS = abi_request_formatters(STANDARD_NORMALIZERS, RPC_ABIS)

FILTER_PARAMS_FORMATTERS = {
    "fromEpoch": to_hex_if_integer,
    "toEpoch": to_hex_if_integer,
    "fromBlock": to_hex_if_integer,
    "toBlock": to_hex_if_integer,
}

PYTHONIC_REQUEST_FORMATTERS: Dict[RPCEndpoint, Callable[..., Any]] = {
    # Cfx

    RPC.cfx_getBalance: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getStakingBalance: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getNextNonce: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getAdmin: apply_formatter_at_index(to_hex_if_integer, 1),
    
    RPC.cfx_getBlockByEpochNumber: apply_formatter_at_index(to_hex_if_integer, 0),
    RPC.cfx_getBlockByBlockNumber: apply_formatter_at_index(to_hex_if_integer, 0),
    RPC.cfx_getBlocksByEpoch: apply_formatter_at_index(to_hex_if_integer, 0),
    RPC.cfx_getSkippedBlocksByEpoch: apply_formatter_at_index(to_hex_if_integer, 0),
    RPC.cfx_getBlockByHashWithPivotAssumption: apply_formatter_at_index(to_hex_if_integer, 2),
    RPC.cfx_getEpochReceipts: apply_formatter_at_index(to_hex_if_integer, 0),
    RPC.cfx_feeHistory: compose(
        apply_formatter_at_index(to_hex_if_integer, 0),
        apply_formatter_at_index(to_hex_if_integer, 1),
    ),
    
    RPC.cfx_getCode: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getStorageAt: apply_formatter_at_index(to_hex_if_integer, 2),
    RPC.cfx_getStorageRoot: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getCollateralForStorage: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getSponsorInfo: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getAccount: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getDepositList: apply_formatter_at_index(to_hex_if_integer, 1),
    RPC.cfx_getVoteList: apply_formatter_at_index(to_hex_if_integer, 1),
    
    RPC.cfx_getInterestRate: apply_formatter_at_index(to_hex_if_integer, 0),
    RPC.cfx_getAccumulateInterestRate: apply_formatter_at_index(to_hex_if_integer, 0),
    RPC.cfx_getBlockRewardInfo: apply_formatter_at_index(to_hex_if_integer, 0),
    RPC.cfx_getPoSEconomics: apply_formatter_at_index(to_hex_if_integer, 0),
    RPC.cfx_getPoSRewardByEpoch: apply_formatter_at_index(to_hex_if_integer, 0),
    RPC.cfx_getParamsFromVote: apply_formatter_at_index(to_hex_if_integer, 0),
    # RPC.cfx_getSupplyInfo: ,
    # RPC.cfx_getAccountPendingInfo: ,
    RPC.cfx_getAccountPendingTransactions: compose(
        apply_formatter_at_index(to_hex_if_integer, 1),
        apply_formatter_at_index(to_hex_if_integer, 2),
    ),
    RPC.cfx_checkBalanceAgainstTransaction: compose(
        apply_formatter_at_index(to_hex_if_integer, 2),
        apply_formatter_at_index(to_hex_if_integer, 3),
        apply_formatter_at_index(to_hex_if_integer, 4),
        apply_formatter_at_index(to_hex_if_integer, 5),
    ),

    RPC.cfx_getLogs: apply_formatter_at_index(apply_formatters_to_dict(FILTER_PARAMS_FORMATTERS), 0),
    RPC.cfx_newFilter: apply_formatter_at_index(apply_formatters_to_dict(FILTER_PARAMS_FORMATTERS), 0),

    RPC.cfx_call: apply_one_of_formatters((
        (is_length(1), apply_formatter_at_index(transaction_param_formatter, 0)), # type: ignore
        (is_length(2), apply_formatters_to_sequence( # type: ignore
            [
                transaction_param_formatter,
                to_hex_if_integer,
            ]
        )),
    )),
    RPC.cfx_estimateGasAndCollateral: apply_one_of_formatters((
        (is_length(1), apply_formatter_at_index(transaction_param_formatter, 0)), # type: ignore
        (is_length(2), apply_formatters_to_sequence( # type: ignore
            [
                transaction_param_formatter,
                to_hex_if_integer,
            ]
        )),
    )),
    RPC.cfx_sendTransaction: apply_formatter_at_index(transaction_param_formatter, 0),
    # RPC.cfx_signTransaction: apply_formatter_at_index(transaction_param_formatter, 0),
    # RPC.eth_getProof: apply_formatter_at_index(to_hex_if_integer, 2),
    # # personal
    # RPC.personal_importRawKey: apply_formatter_at_index(
    #     compose(remove_0x_prefix, hexstr_if_str(to_hex)),
    #     0,
    # ),
    # RPC.personal_sign: apply_formatter_at_index(text_if_str(to_hex), 0),
    # RPC.personal_ecRecover: apply_formatter_at_index(text_if_str(to_hex), 0),
    # RPC.personal_sendTransaction: apply_formatter_at_index(transaction_param_formatter, 0),
    # # Snapshot and Revert
    # RPC.evm_revert: apply_formatter_at_index(integer_to_hex, 0),
    # RPC.trace_replayBlockTransactions: apply_formatter_at_index(to_hex_if_integer, 0),
    # RPC.trace_block: apply_formatter_at_index(to_hex_if_integer, 0),
    # RPC.trace_call: compose(
    #     apply_formatter_at_index(transaction_param_formatter, 0),
    #     apply_formatter_at_index(to_hex_if_integer, 2)
    # ),
}

STATUS_FORMATTERS = {
    "bestHash": to_hash32,
    "chainId": to_integer_if_hex,
    "networkId": to_integer_if_hex,
    "blockNumber": to_integer_if_hex,
    "epochNumber": to_integer_if_hex,
    "latestCheckpoint": to_integer_if_hex,
    "latestConfirmed": to_integer_if_hex,
    "latestState": to_integer_if_hex,
    "latestFinalized": to_integer_if_hex,
    "ethereumSpaceChainId": to_integer_if_hex,
    "pendingTxNumber": to_integer_if_hex
}

ESTIMATE_FORMATTERS = {
    "gasLimit": to_integer_if_hex,
    "gasUsed": to_integer_if_hex,
    "storageCollateralized": to_integer_if_hex
}


LOG_ENTRY_FORMATTERS = {
    "address": from_trust_to_base32,
    "topics": apply_list_to_array_formatter(to_hash32), 
    "data": HexBytes,
    "blockHash": apply_formatter_if(is_not_null, to_hash32), 
    "epochNumber": apply_formatter_if(is_not_null, to_integer_if_hex),
    "transactionHash": apply_formatter_if(is_not_null, to_hash32), 
    "transactionIndex": apply_formatter_if(is_not_null, to_integer_if_hex),
    "logIndex": apply_formatter_if(is_not_null, to_integer_if_hex),
    "transactionLogIndex": apply_formatter_if(is_not_null, to_integer_if_hex),
}
log_entry_formatter = apply_formatters_to_dict(LOG_ENTRY_FORMATTERS)

STORAGE_CHANGE_FORMATTER = apply_formatters_to_dict({
    "address": from_trust_to_base32,
    "collaterals": to_integer_if_hex,
})

RECEIPT_FORMATTERS = {
    "type": to_integer_if_hex,
    "transactionHash": to_hash32, 
    "index": to_integer_if_hex,
    "blockHash": apply_formatter_if(is_not_null, to_hash32), 
    "epochNumber": to_integer_if_hex,
    "from": from_trust_to_base32,
    "to": apply_formatter_if(is_not_null, from_trust_to_base32),
    "gasUsed": to_integer_if_hex,
    "gasFee": from_hex_to_drip,
    # "gasCoveredBySponsor": bool,
    "storageCollateralized": to_integer_if_hex,
    # "storageCoveredBySponsor": bool,
    "storageReleased": apply_list_to_array_formatter(STORAGE_CHANGE_FORMATTER),
    "contractCreated": apply_formatter_if(is_not_null, from_trust_to_base32),
    
    "stateRoot": to_hash32,
    "outcomeStatus": to_integer_if_hex,
    "logsBloom": to_hexbytes(256), # type: ignore
    "logs": apply_list_to_array_formatter(log_entry_formatter),
    "burntGasFee": from_hex_to_drip,
    "effectiveGasPrice": from_hex_to_drip,
}
# receipt_formatter = apply_formatters_to_dict(RECEIPT_FORMATTERS)

ACCESS_LIST_ENTRY_FORMATTERS = {
    "address": from_trust_to_base32,
}

TRANSACTION_DATA_FORMATTERS = {
    "type": to_integer_if_hex,
    "blockHash": apply_formatter_if(is_not_null, to_hash32),
    "chainId": apply_formatter_if(is_not_null, to_integer_if_hex),
    "contractCreated": apply_formatter_if(is_not_null, from_trust_to_base32),
    "data": HexBytes,
    "epochHeight": apply_formatter_if(is_not_null, to_integer_if_hex),
    "from": from_trust_to_base32,
    "gas": to_integer_if_hex,
    "gasPrice": from_hex_to_drip,
    "hash": to_hash32, # 
    "nonce": to_integer_if_hex,
    "r": apply_formatter_if(is_not_null, to_hexbytes(32, variable_length=True)), # type: ignore
    "s": apply_formatter_if(is_not_null, to_hexbytes(32, variable_length=True)), # type: ignore
    "status": to_integer_if_hex,
    "storageLimit": to_integer_if_hex,
    "to": apply_formatter_if(is_not_null, from_trust_to_base32),
    "transactionIndex": apply_formatter_if(is_not_null, to_integer_if_hex),
    "v": apply_formatter_if(is_not_null, to_integer_if_hex),
    "value": from_hex_to_drip,
    "maxFeePerGas": apply_formatter_if(is_not_null, from_hex_to_drip),
    "maxPriorityFeePerGas": apply_formatter_if(is_not_null, from_hex_to_drip),
    "accessList": apply_formatter_if(is_not_null, apply_list_to_array_formatter(apply_formatters_to_dict(ACCESS_LIST_ENTRY_FORMATTERS))),
    "yParity": apply_formatter_if(is_not_null, to_integer_if_hex),
}
transaction_data_formatter = apply_formatters_to_dict(TRANSACTION_DATA_FORMATTERS)

filter_result_formatter = apply_one_of_formatters(
    (
        (is_array_of_dicts, apply_list_to_array_formatter(log_entry_formatter)),
        (is_array_of_strings, apply_list_to_array_formatter(to_hash32)), 
    )
)

BLOCK_FORMATTERS = {
    "hash": to_hash32,
    "parentHash": to_hash32,
    "height": to_integer_if_hex,
    "miner": from_trust_to_base32,
    "deferredStateRoot": to_hash32,
    "deferredReceiptsRoot": to_hash32,
    "deferredLogsBloomHash": to_hash32,
    "blame": to_integer_if_hex,
    "transactionsRoot": to_hash32,
    "epochNumber": apply_formatter_if(is_not_null, to_integer_if_hex),
    "blockNumber": apply_formatter_if(is_not_null, to_integer_if_hex),
    "gasLimit": to_integer_if_hex,
    "gasUsed": apply_formatter_if(is_not_null, to_integer_if_hex),
    "timestamp": to_integer_if_hex,
    "difficulty": to_integer_if_hex,
    "powQuality": apply_formatter_if(is_not_null, HexBytes),
    "refereeHashes": apply_list_to_array_formatter(to_hash32),
    # adaptive: bool
    "nonce": apply_formatter_if(is_not_null, to_hexbytes(8, variable_length=True)), # type: ignore
    "size": to_integer_if_hex,
    "custom": apply_list_to_array_formatter(HexBytes),
    "posReference": apply_formatter_if(is_not_null, to_hash32), # set for development
    "transactions": apply_one_of_formatters(
        (
            (
                is_array_of_dicts,
                apply_list_to_array_formatter(transaction_data_formatter),
            ),
            (is_array_of_strings, apply_list_to_array_formatter(to_hash32)),
        )
    ),
    "baseFeePerGas": apply_formatter_if(is_not_null, from_hex_to_drip),
}
block_formatter = apply_formatters_to_dict(BLOCK_FORMATTERS)


def to_transaction_hash(val: Hash32) -> TransactionHash:
    if isinstance(val, TransactionHash):
        return val
    return TransactionHash(to_hexbytes(32, val))

def fixed64_to_float(val: str) -> float:
    MAX = 2**256 - 1
    return int(val, 16) / MAX

STORAGE_ROOT_FORMATTERS = {
    "delta": apply_formatter_if(is_hash32_str, to_hash32),
    "intermediate": apply_formatter_if(is_hash32_str, to_hash32),
    "snapshot": apply_formatter_if(is_hash32_str, to_hash32),
}
storage_root_formatter = apply_formatters_to_dict(STORAGE_ROOT_FORMATTERS)

SPONSOR_INFO_FORMATTERS = {
    "sponsorBalanceForCollateral": from_hex_to_drip,
    "sponsorBalanceForGas": from_hex_to_drip,
    "sponsorGasBound": from_hex_to_drip,
    "sponsorForCollateral": from_trust_to_base32,
    "sponsorForGas": from_trust_to_base32,
    "availableStoragePoints": to_integer_if_hex,
    "usedStoragePoints": to_integer_if_hex,
}

ACCOUNT_INFO_FORMATTERS = {
    "accumulatedInterestReturn": from_hex_to_drip,
    "address": from_trust_to_base32,
    "admin": apply_formatter_if(is_not_null, from_trust_to_base32),
    "balance": from_hex_to_drip,
    "codeHash": to_hash32,
    "collateralForStorage": to_integer_if_hex,
    "nonce": to_integer_if_hex,
    "stakingBalance": from_hex_to_drip,
}

DEPOSIT_INFO_FORMATTERS = {
    "accumulatedInterestRate": to_integer_if_hex,
    "amount": from_hex_to_drip,
    "depositTime": to_integer_if_hex,
}

VOTE_INFO_FORMATTERS = {
    "amount": from_hex_to_drip,
    "unlockBlockNumber": to_integer_if_hex,
}

BLOCK_REWARD_INFO_FORMATTERS = {
    "blockHash": to_hash32,
    "author": from_trust_to_base32,
    "totalReward": from_hex_to_drip,
    "baseReward": from_hex_to_drip,
    "txFee": from_hex_to_drip,
}

POS_ECONOMICS_FORMATTERS = {
    "distributablePosInterest": from_hex_to_drip,
    "lastDistributeBlock": to_integer_if_hex,
    "totalPosStakingTokens": from_hex_to_drip,
}

POS_ACCOUNT_REWARDS_FORMATTERS = {
    # "posAddress": hexaddress
    "powAddress": from_trust_to_base32,
    "reward": from_hex_to_drip
}

POS_REWARDS_INFO_FORMATTERS = {
    "accountRewards": apply_list_to_array_formatter(
        apply_formatters_to_dict(POS_ACCOUNT_REWARDS_FORMATTERS)
    ),
    "powEpochHash": to_hash32
}

DAO_INFO_FORMATTERS = {
    "powBaseReward": from_hex_to_drip,
    "interestRate": to_integer_if_hex,
    "storagePointProp": to_integer_if_hex,
    "baseFeeShareProp": to_integer_if_hex,
}

SUPPLY_INFO_FORMATTERS = {
    "totalCirculating": from_hex_to_drip,
    "totalCollateral": from_hex_to_drip,
    "totalIssued": from_hex_to_drip,
    "totalStaking": from_hex_to_drip,
    "totalEspaceTokens": from_hex_to_drip,
}

PENDING_INFO_FORMATTERS = {
    "localNonce": to_integer_if_hex,
    "nextPendingTx": to_hash32,
    "pendingCount": to_integer_if_hex,
    "pendingNonce": to_integer_if_hex,
}

PENDING_TRANSACTIONS_INFO_FORMATTERS = {
    "pendingCount": to_integer_if_hex,
    "pendingTransactions": apply_list_to_array_formatter(
        transaction_data_formatter
    )
}

FEE_HISTORY_FORMATTERS = {
    "baseFeePerGas": apply_list_to_array_formatter(from_hex_to_drip),
    # "gasUsedRatio": lambda x: x, # do nothing
    "oldestEpoch": to_integer_if_hex,
    "reward": apply_list_to_array_formatter(apply_list_to_array_formatter(from_hex_to_drip)),
}


SIMPLE_RESULT_FORMATTER_MAPPING: Dict[Type[Any], Callable[..., Any]] = {
    int: to_integer_if_hex,
    Drip: from_hex_to_drip,
    Base32Address: from_trust_to_base32,
    Hash32: to_hash32,
    HexBytes: HexBytes,
    Union[int, None]: apply_formatter_if(is_not_null, to_integer_if_hex)
}

def create_dict_result_formatter(typed_dict_class: Type[TypedDict]) -> Callable[[Dict[str, Any]], Dict[str, Any]]:
    formatter_dict: Dict[str, Callable[..., Any]] = {}
    type_hints = get_type_hints(typed_dict_class)
    
    for key, type_hint in type_hints.items():
        # Get the formatter function for the specific type hint
        formatter_function = SIMPLE_RESULT_FORMATTER_MAPPING.get(type_hint, None)
        
        if formatter_function is not None:
            formatter_dict[key] = formatter_function
            
    return apply_formatters_to_dict(formatter_dict)

PYTHONIC_RESULT_FORMATTERS: Dict[RPCEndpoint, Callable[..., Any]] = {
    # cfx namespace
    RPC.cfx_epochNumber: to_integer_if_hex,
    RPC.cfx_getStatus: apply_formatters_to_dict(STATUS_FORMATTERS),
    RPC.cfx_call: HexBytes,
    RPC.cfx_estimateGasAndCollateral: apply_formatters_to_dict(ESTIMATE_FORMATTERS),
    RPC.cfx_getConfirmationRiskByHash: fixed64_to_float,
    RPC.cfx_gasPrice: from_hex_to_drip,
    RPC.cfx_maxPriorityFeePerGas: from_hex_to_drip,
    RPC.cfx_getBalance: from_hex_to_drip,
    RPC.cfx_getStakingBalance: from_hex_to_drip,
    RPC.cfx_getNextNonce: to_integer_if_hex,
    RPC.cfx_getBlockByHash: apply_formatter_if(is_not_null, block_formatter),
    RPC.cfx_getBlockByEpochNumber: apply_formatter_if(is_not_null, block_formatter),
    RPC.cfx_getBlockByBlockNumber: apply_formatter_if(is_not_null, block_formatter),
    RPC.cfx_getBestBlockHash: to_hash32,
    RPC.cfx_getBlocksByEpoch: apply_list_to_array_formatter(to_hash32),
    RPC.cfx_getSkippedBlocksByEpoch: apply_list_to_array_formatter(to_hash32),
    RPC.cfx_getBlockByHashWithPivotAssumption: apply_formatter_if(is_not_null, block_formatter),
    RPC.cfx_getEpochReceipts: apply_list_to_array_formatter(apply_list_to_array_formatter(apply_formatter_if(
        is_not_null,
        apply_formatters_to_dict(RECEIPT_FORMATTERS),
    ))),
    RPC.cfx_feeHistory: apply_formatters_to_dict(FEE_HISTORY_FORMATTERS),

    RPC.cfx_getLogs: filter_result_formatter,
    RPC.cfx_getFilterLogs: filter_result_formatter,
    RPC.cfx_getFilterChanges: filter_result_formatter,
    
    RPC.cfx_getCode: HexBytes,
    RPC.cfx_getAdmin: apply_formatter_if(is_not_null, from_trust_to_base32),
    RPC.cfx_getStorageAt: apply_formatter_if(is_not_null, to_hash32),
    RPC.cfx_getStorageRoot: apply_formatter_if(is_not_null, storage_root_formatter),
    RPC.cfx_getCollateralForStorage: to_integer_if_hex,
    RPC.cfx_getSponsorInfo: apply_formatters_to_dict(SPONSOR_INFO_FORMATTERS),
    RPC.cfx_getAccount: apply_formatters_to_dict(ACCOUNT_INFO_FORMATTERS),
    RPC.cfx_getDepositList: apply_list_to_array_formatter(
        apply_formatters_to_dict(DEPOSIT_INFO_FORMATTERS)
    ),
    RPC.cfx_getVoteList: apply_list_to_array_formatter(
        apply_formatters_to_dict(VOTE_INFO_FORMATTERS)
    ),
    
    RPC.cfx_getInterestRate: to_integer_if_hex,
    RPC.cfx_getAccumulateInterestRate: to_integer_if_hex,
    RPC.cfx_getBlockRewardInfo: apply_list_to_array_formatter(
        apply_formatters_to_dict(BLOCK_REWARD_INFO_FORMATTERS)
    ),
    RPC.cfx_getPoSEconomics: apply_formatter_if(
        is_not_null, 
        apply_formatters_to_dict(POS_ECONOMICS_FORMATTERS),
    ),
    RPC.cfx_getPoSRewardByEpoch: apply_formatter_if(
        is_not_null,
        apply_formatters_to_dict(POS_REWARDS_INFO_FORMATTERS),
    ),
    RPC.cfx_getParamsFromVote: apply_formatter_if(
        is_not_null,
        apply_formatters_to_dict(DAO_INFO_FORMATTERS)
    ),
    RPC.cfx_getSupplyInfo: apply_formatters_to_dict(SUPPLY_INFO_FORMATTERS),
    RPC.cfx_getCollateralInfo: create_dict_result_formatter(CollateralInfo),
    RPC.cfx_getFeeBurnt: from_hex_to_drip,

    RPC.cfx_getAccountPendingInfo: apply_formatters_to_dict(PENDING_INFO_FORMATTERS),
    RPC.cfx_getAccountPendingTransactions: apply_formatters_to_dict(PENDING_TRANSACTIONS_INFO_FORMATTERS),
    RPC.cfx_getTransactionByHash: apply_formatter_if(
        is_not_null,
        transaction_data_formatter
    ),
    # RPC.eth_getTransactionCount: to_integer_if_hex,
    RPC.cfx_getTransactionReceipt: apply_formatter_if(
        is_not_null,
        apply_formatters_to_dict(RECEIPT_FORMATTERS),
    ),
    
    RPC.cfx_sendRawTransaction: to_transaction_hash,  
    RPC.cfx_sendTransaction: to_transaction_hash,  
    # RPC.eth_sign: HexBytes,
    # RPC.eth_signTransaction: apply_formatter_if(is_not_null, signed_tx_formatter),
    # RPC.eth_signTypedData: HexBytes,
    # RPC.eth_syncing: apply_formatter_if(is_not_false, syncing_formatter),
    # # personal
    # RPC.personal_importRawKey: to_checksum_address,
    # RPC.personal_listAccounts: apply_list_to_array_formatter(to_checksum_address),
    # RPC.personal_listWallets: apply_list_to_array_formatter(geth_wallets_formatter),
    # RPC.personal_newAccount: to_checksum_address,
    # RPC.personal_sendTransaction: to_hash32,
    # RPC.personal_signTypedData: HexBytes,
    
    # Transaction Pool
    RPC.txpool_nextNonce: to_integer_if_hex,
}


def cfx_request_formatters(
    method_name: Union[RPCEndpoint, Callable[..., RPCEndpoint]]
) -> Dict[str, Callable[..., Any]]:
    request_formatter_maps = (
        ABI_REQUEST_FORMATTERS,
        # METHOD_NORMALIZERS needs to be after ABI_REQUEST_FORMATTERS
        # so that eth_getLogs's apply_formatter_at_index formatter
        # is applied to the whole address
        # rather than on the first byte of the address
        # METHOD_NORMALIZERS,
        PYTHONIC_REQUEST_FORMATTERS,
    )
    formatters = combine_formatters(request_formatter_maps, method_name)
    return compose(*formatters)


def is_attrdict(val: Any) -> bool:
    return isinstance(val, AttributeDict)
not_attrdict = complement(is_attrdict)

def cfx_result_formatters(
    method_name: Union[RPCEndpoint, Callable[..., RPCEndpoint]],
    module: "Module",
) -> Dict[str, Callable[..., Any]]:
    formatters = combine_formatters((PYTHONIC_RESULT_FORMATTERS,), method_name)
    # formatters_requiring_module = combine_formatters(
    #     (FILTER_RESULT_FORMATTERS,), method_name
    # )

    # partial_formatters = apply_module_to_formatters(
    #     formatters_requiring_module, module, method_name
    # )
    attrdict_formatter = apply_formatter_if(
        is_dict and not_attrdict, AttributeDict.recursive
    )
    return compose(attrdict_formatter, *formatters)
